/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2009 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver.h,v 1.109 2009/01/09 01:43:10 zlb Exp $ */

#ifndef PHG_SOLVER_H

/* The SOLVER struct */
typedef struct SOLVER_ {
    FLOAT	residual;	/* residual or error norm (filled by Solve() */
    FLOAT       rtol;		/* relative convergence tolerance */
    FLOAT       atol;		/* absolute tolerance */
    FLOAT       btol;		/* b-tolerance */
    INT		nits;		/* number of iterations (filled by Solve() */
    INT		maxit;		/* Maxi. number of iterations for oem_solver */
    MAT		*mat;		/* the matrix */
    VEC		*rhs;		/* local right-hand-side vector */
    struct OEM_SOLVER_	*oem_solver;
    void		*oem_data;
    struct SOLVER_PC_	*pc;		/* PreConditioner for solver */

    BOOLEAN	rhs_updated;	/* TRUE ==> RHS updated with bdry eqns */
    BOOLEAN	sent_to_oem;	/* TRUE ==> data sent to OEM solver */
    BOOLEAN	oem_created;	/* TRUE ==> oem_solver->Create() called */
    BOOLEAN     warn_maxit;	/* TRUE ==> warn when maxit achieved */
} SOLVER;

/* The OEM_SOLVER struct */
typedef struct OEM_SOLVER_ {
    char *name;					/* solver name */
    int (*RegisterOptions)(void);
    int (*Initialize)(int *argc, char ***argv); /* by phgSolverInitialize */
    int (*Finalize)(void);

    int (*Init)(SOLVER *solver);	/* called by phgSolverCreate */
    int (*Create)(SOLVER *solver);	/* called by phgSolverSolve */
    int (*Destroy)(SOLVER *solver);
    int (*AddMatrixEntries)(SOLVER *solver, INT nrows, INT *rows,
		INT ncols, INT *cols, FLOAT *values);
    int (*AddRHSEntries)(SOLVER *solver, INT n, INT *ni, FLOAT *values);
    int (*Assemble)(SOLVER *solver);
    int (*Solve)(SOLVER *solver, VEC *x, BOOLEAN destroy);
    BOOLEAN iterative;			/* whether an iterative solver */
    BOOLEAN parallel;			/* whether a parallel solver */
} OEM_SOLVER;

/* pre-conditioner */
/* preconditioner interface */
typedef void (*PC_PROC)(SOLVER *pc_solver, VEC *b, VEC **x);
/* SOLVER_PC struct for builtin solver's preconditioner  */
typedef struct SOLVER_PC_ {
    char *name;
    SOLVER	*solver;
    PC_PROC	pc_proc;
    BOOLEAN	owns_solver;
} SOLVER_PC;

/* export phgSolverNames and phgSolverList to user programs */
extern const char *phgSolverNames_[];
extern OEM_SOLVER *phgSolverList_[];
#define phgSolverNames	(phgSolverNames_ + 1)
#define phgSolverList	(phgSolverList_ + 1)

/* default solver */
#define SOLVER_DEFAULT	((OEM_SOLVER *)NULL)	/* default solver */

/* PETSc solver */
#if USE_PETSC
extern OEM_SOLVER phgSolverPetsc_;
#define SOLVER_PETSC	(&phgSolverPetsc_)
#else
#define SOLVER_PETSC	((OEM_SOLVER *)NULL)
#endif

/* Trilinos solver */
#if USE_TRILINOS
extern OEM_SOLVER phgSolverTrilinos_;
#define SOLVER_TRILINOS	(&phgSolverTrilinos_)
#else
#define SOLVER_TRILINOS	((OEM_SOLVER *)NULL)
#endif

/* SPC solver */
#if USE_SPC
extern OEM_SOLVER phgSolverSPC_;
#define SOLVER_SPC	(&phgSolverSPC_)
#else
#define SOLVER_SPC	((OEM_SOLVER *)NULL)
#endif

/* SuperLU solver */
#if USE_SUPERLU
extern OEM_SOLVER phgSolverSuperLU_;
#define SOLVER_SUPERLU	(&phgSolverSuperLU_)
#else
#define SOLVER_SUPERLU	((OEM_SOLVER *)NULL)
#endif

/* MUMPS solver */
#if USE_MUMPS
extern OEM_SOLVER phgSolverMUMPS_;
#define SOLVER_MUMPS	(&phgSolverMUMPS_)
#else
#define SOLVER_MUMPS	((OEM_SOLVER *)NULL)
#endif

/* X9AMG solver */
#if USE_X9AMG
extern OEM_SOLVER phgSolverX9AMG_;
#define SOLVER_X9AMG	(&phgSolverX9AMG_)
#else
#define SOLVER_X9AMG	((OEM_SOLVER *)NULL)
#endif

/* SPOOLES solver */
#if USE_SPOOLES
extern OEM_SOLVER phgSolverSPOOLES_;
#define SOLVER_SPOOLES	(&phgSolverSPOOLES_)
#else
#define SOLVER_SPOOLES	((OEM_SOLVER *)NULL)
#endif

/* PARDISO solver */
#if USE_PARDISO
extern OEM_SOLVER phgSolverPARDISO_;
#define SOLVER_PARDISO	(&phgSolverPARDISO_)
#else
#define SOLVER_PARDISO	((OEM_SOLVER *)NULL)
#endif

/* SuiteSparse solvers */
#if USE_SSPARSE
extern OEM_SOLVER phgSolverSSPARSE_;
#define SOLVER_SSPARSE	(&phgSolverSSPARSE_)
#else
#define SOLVER_SSPARSE	((OEM_SOLVER *)NULL)
#endif

/* LASPack solver */
#if USE_LASPACK
extern OEM_SOLVER phgSolverLASPack_;
#define SOLVER_LASPACK	(&phgSolverLASPack_)
#else
#define SOLVER_LASPACK	((OEM_SOLVER *)NULL)
#endif

/* HYPRE solver */
#if USE_HYPRE
extern OEM_SOLVER phgSolverHYPRE_;
#define SOLVER_HYPRE	(&phgSolverHYPRE_)
#else
#define SOLVER_HYPRE	((OEM_SOLVER *)NULL)
#endif

/* Built-in solver */
extern OEM_SOLVER phgSolverBuiltIn_;
#define SOLVER_BUILTIN	(&phgSolverBuiltIn_)

/* Built-in solver */
extern OEM_SOLVER phgSolverPCG_;
#define SOLVER_PCG	(&phgSolverPCG_)

extern OEM_SOLVER phgSolverGMRES_;
#define SOLVER_GMRES	(&phgSolverGMRES_)

extern OEM_SOLVER phgSolverAMS_;
#define SOLVER_AMS	(&phgSolverAMS_)

#ifdef __cplusplus
extern "C" {
#endif

#define phgSolverMapD2L(solver, dof_no, index) \
	phgMapD2L((solver)->mat->cmap, dof_no, index)

#define phgSolverMapE2L(solver, dof_no, e, index) \
	phgMapE2L((solver)->mat->cmap, dof_no, e, index)

#define phgSolverMapL2G(solver, index) \
	phgMapL2G((solver)->mat->cmap, index)

#define phgSolverMapE2G(solver, dof_no, e, index) \
	phgMapE2G((solver)->mat->cmap, dof_no, e, index)

/* functions provided by solver interface */
int phgSolverRegisterOptions(void);
int phgSolverInitialize(int *argc, char ***argv);
int phgSolverFinalize(void);

SOLVER *phgSolverMat2Solver(OEM_SOLVER *oem_solver, MAT *mat);
MAT *phgSolverGetMat(SOLVER *solver);
SOLVER *phgSolverCreate(OEM_SOLVER *oem_solver, DOF *u, ...);
int phgSolverDestroy(SOLVER **solver);

#define phgSolverAddMatrixEntry(solver, row, col, value) \
	phgMatAddEntry((solver)->mat, row, col, value)
#define phgSolverAddGlobalMatrixEntry(solver, row, col, value) \
	phgMatAddGlobalEntry((solver)->mat, row, col, value)

#define phgSolverAddMatrixEntries(solver, nrows, rows, ncols, cols, values) \
	phgMatAddEntries((solver)->mat, nrows, rows, ncols, cols, values)
#define phgSolverAddGlobalMatrixEntries(solver,nrows,rows,ncols,cols,values) \
	phgMatAddGlobalEntries((solver)->mat, nrows, rows, ncols, cols, values)

#define phgSolverAddRHSEntry(solver, index, value) \
	phgVecAddEntry((solver)->rhs, 0, index, value)
#define phgSolverAddGlobalRHSEntry(solver, index, value) \
	phgVecAddGlobalEntry((solver)->rhs, 0, index, value)

#define phgSolverAddRHSEntries(solver, n, indices, values) \
	phgVecAddEntries((solver)->rhs, 0, n, indices, values)
#define phgSolverAddGlobalRHSEntries(solver, n, indices, values) \
	phgVecAddGlobalEntries((solver)->rhs, 0, n, indices, values)

BOOLEAN phgSolverDenseLU(int n, FLOAT *a, int pvt[]);
void phgSolverDenseSV(int n, FLOAT *a, int pvt[], int m, FLOAT *b);
BOOLEAN phgSolverDenseSolver(int n, int m, FLOAT *a, FLOAT *b);

#define phgSolverAssemble(solver)	phgMatAssemble((solver)->mat)
#define phgSolverAssembleRHS(solver)	phgVecAssemble((solver)->rhs)

void phgSolverUpdateRHS(SOLVER *solver);
int phgSolverSolve(SOLVER *solver, BOOLEAN destroy, DOF *u, ...);
void phgSolverSetMaxIt(SOLVER *solver, int maxit);
void phgSolverSetTol(SOLVER *solver, FLOAT user_tol);

int phgSolverResetRHS(SOLVER *solver);
int phgSolverVecSolve(SOLVER *solver, BOOLEAN destroy, VEC *x);

void phgSolverSetPC(SOLVER *solver, SOLVER *pc_solver, PC_PROC pc_func);

void phgSolverAMSRemoveGradient(DOF **fptr, DOF *beta, int order);
void phgSolverAMSSetCoefficients(SOLVER *solver, DOF *alpha, DOF *beta);
void phgSolverAMSSetConstantCoefficients(SOLVER *solver, FLOAT a, FLOAT b);

#if USE_HYPRE && HYPRE_VERSION_MAJOR >= 2
void phgSolverHypreAMSSetPoisson_(SOLVER *solver, MAT *alpha, MAT *beta,
				  BOOLEAN destroy);
void phgSolverHypreAMSSetPoisson(SOLVER *solver, DOF *alpha, DOF *beta);
void phgSolverHypreAMSSetConstantPoisson(SOLVER *solver,
						FLOAT alpha, FLOAT beta);
#else	/* USE_HYPRE && HYPRE_VERSION_MAJOR >= 2 */
# define phgSolverHypreAMSSetPoisson(solver, a, b)	    /* empty macro */
# define phgSolverHypreAMSSetConstantPoisson(solver, a, b)  /* empty macro */
#endif	/* USE_HYPRE && HYPRE_VERSION_MAJOR >= 2 */

#ifdef __cplusplus
}
#endif

#define PHG_SOLVER_H
#endif
