/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2009 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: partition-sfc.c,v 1.49 2008/12/26 08:29:46 zlb Exp $ */

/* SFC Space filling curve partitioner */
/* Hilbert and Morton methods */
#include "phg.h"
#include "partition-sfc.h"

#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

/* **************************************************************
 * Inverse Hilbert Space-Filling curve 
 * converte coordinates in [0,1]x[0,1]x[0,1] to [0, 1]
 * n <= the length of array x 
 * 
 * Here we give two versions, the first one is general method
 * that based on table, which is based on Zoltan's.
 * We extented original implement, and we fixed some bugs.
 * The second one is based on an improved algorithm, which we
 * modified the first one. The time complexity is O(r), 
 * r = log2(max(x, y, z)) + 1. In general, r < n;
 * Both are fast. User can choose any one.
 * If grid has big coordinates much more than small coordinates, the 
 *   first maybe better.
 * If grid has small coordinats much more than bigs ones, the second
 *   maybe better.
 * **************************************************************/
/*  maximum level */
/*  in this file, we force that MAXLEVEL >= 20 && MAXLEVEL <= 30 */
#define  MAXLEVEL  25

/* first algorithm */
#if 1

BOOLEAN
phgInvHilbertSFC(SFC *x, INT n)
{
    static const UINT *d[] =
        {idata3d,       idata3d + 8,   idata3d + 16,  idata3d + 24,
         idata3d + 32,  idata3d + 40,  idata3d + 48,  idata3d + 56,
         idata3d + 64,  idata3d + 72,  idata3d + 80,  idata3d + 88,
         idata3d + 96,  idata3d + 104, idata3d + 112, idata3d + 120,
         idata3d + 128, idata3d + 136, idata3d + 144, idata3d + 152,
         idata3d + 160, idata3d + 168, idata3d + 176, idata3d + 184
        };

    static const UINT *s[] =
        {istate3d,       istate3d + 8,   istate3d + 16,  istate3d + 24,
         istate3d + 32,  istate3d + 40,  istate3d + 48,  istate3d + 56,
         istate3d + 64,  istate3d + 72,  istate3d + 80,  istate3d + 88,
         istate3d + 96,  istate3d + 104, istate3d + 112, istate3d + 120,
         istate3d + 128, istate3d + 136, istate3d + 144, istate3d + 152,
         istate3d + 160, istate3d + 168, istate3d + 176, istate3d + 184
        };

    INT level, EL;
    UINT key[3], c[3], temp, state;
    INT i;

    static UINT  IMAX;
    static UINT EfBit;
    static BOOLEAN initialized = FALSE;
    static INT k0 = 0, k1 = 0, k2 = 0;

    if (!initialized) {
        initialized = TRUE;

        /* Init IMAX such that all effective bits are 1 */
        /* IMAX = 2^32 - 1 */
        IMAX = 4294967295U;

        /* 30 effective bits, all 1 */
        /* EfBit = 2^30 - 1 */
        EfBit = IMAX >> 2;

        k0 = 60 - MAXLEVEL * 3;
        k1 = 30 - MAXLEVEL * 3;
        k2 = - MAXLEVEL * 3;
    }

    for (i = 0; i < n; i++) {
        /* verification coordinats */
#if 0
        assert(x[i].co[0] >= 0. && x[i].co[0] <= 1.);
        assert(x[i].co[1] >= 0. && x[i].co[1] <= 1.);
        assert(x[i].co[2] >= 0. && x[i].co[2] <= 1.);
#endif
        /* convert x,y,z coordinates to integers in range [0,IMAX] */
        c[0] = (UINT)(x[i].co[0] * (double)IMAX);        /* x */
        c[1] = (UINT)(x[i].co[1] * (double)IMAX);        /* y */
        c[2] = (UINT)(x[i].co[2] * (double)IMAX);        /* z */
        c[1] >>= 1;
        c[2] >>= 2;

        /* use state tables to convert nested quadrant's 
         * coordinates level by level */
        key[0] = key[1] = key[2] = 0;
        state = 0;
        EL = 30;
        for (level = 0; level < MAXLEVEL; level++) {
            /* extract 3 bits at current level */
            EL--;
            temp = ((c[0] >> EL) & 4)
                  |((c[1] >> EL) & 2)
                  |((c[2] >> EL) & 1);

            /* treat key[] as long shift register */
            /* shift in converted coordinate */
            /* every key[] has thirty effective bit */
            /* the last 30 are effective */
            key[0] = (key[0] << 3) | ((key[1] >> 27) & 7);
            key[1] = (key[1] << 3) | ((key[2] >> 27) & 7);
            key[2] = (key[2] << 3) | *(d[state] + temp);

            state = *(s[state] + temp);
        }
        
        key[0] = key[0] & EfBit;
        key[1] = key[1] & EfBit;
        key[2] = key[2] & EfBit;

        /* convert 2 part Hilbert key to double */
        x[i].sfc =  ldexp((double)key[0], k0) 
                  + ldexp((double)key[1], k1);
                  + ldexp((double)key[2], k2);
    }

    return (i == n);
}
#else
/* log2(n)             */
/* n >= 0              */
/* return 0, if n <= 1 */
static INT
LOG2(INT n)
{
    INT lo, hi, md;

    if (n <= 1) {
        return 0;
    }

    lo = 1;
    hi = MAXLEVEL;
    md = (lo + hi) >> 1;

    while(1) {
        if ((n >> md) > 0) {
            lo = md;
        }
        else {
            hi = md;
        }

        md = (lo + hi) >> 1;
        if (md == lo) {
            break;
        }
    }

    return lo;
}

BOOLEAN
phgInvHilbertSFC(SFC * x, INT n)
{
    /* ordering data */
    static const UINT *d[] =
        {idata3d,       idata3d + 8,   idata3d + 16,  idata3d + 24,
         idata3d + 32,  idata3d + 40,  idata3d + 48,  idata3d + 56,
         idata3d + 64,  idata3d + 72,  idata3d + 80,  idata3d + 88,
         idata3d + 96,  idata3d + 104, idata3d + 112, idata3d + 120,
         idata3d + 128, idata3d + 136, idata3d + 144, idata3d + 152,
         idata3d + 160, idata3d + 168, idata3d + 176, idata3d + 184
        };
    /* orientation(status) data */
    static const UINT *s[] =
        {istate3d,       istate3d + 8,   istate3d + 16,  istate3d + 24,
         istate3d + 32,  istate3d + 40,  istate3d + 48,  istate3d + 56,
         istate3d + 64,  istate3d + 72,  istate3d + 80,  istate3d + 88,
         istate3d + 96,  istate3d + 104, istate3d + 112, istate3d + 120,
         istate3d + 128, istate3d + 136, istate3d + 144, istate3d + 152,
         istate3d + 160, istate3d + 168, istate3d + 176, istate3d + 184
        };
    /* auxiliary parameters, constants */
    static UINT  IMAX;
    static UINT EfBit;
    static BOOLEAN initialized = FALSE;
    static INT k0, k1, k2;

    INT level, EL;
    UINT key[3], c[3], temp, state;
    INT i, r;

    /* Initialize Inverse SFC parameters */
    if (!initialized) {
        if (phgVerbosity > 0) {
            phgPrintf("Initialize Inverse SFC parameters\n");
        }

        initialized = TRUE;

        /* Init IMAX such that all effective bits are 1 */
        /* IMAX = 2^(MAXLEVEL + 2) - 1 */
        /* Will be reduced later */
        IMAX = 1;
        for (i = 0; i < MAXLEVEL; i++) {
            IMAX <<= 1;
        }
        IMAX -= 1;

        /* 30 effective bits, all 1 */
        /* EfBit = 2^30 - 1 */
        EfBit = 1;
        for (i = 0; i < 30; i++) {
            EfBit <<= 1;
        }
        EfBit -= 1;

        /* index used by ldexp */
        /* effective number */
        k0 = 60 - MAXLEVEL * 3;
        k1 = 30 - MAXLEVEL * 3;
        k2 = - MAXLEVEL * 3;

    } /* end: initialize */

    for (i = 0; i < n; i++) {
        /* verificate coordinats */
#if 0
        assert(x[i].co[0] >= 0. && x[i].co[0] <= 1.);
        assert(x[i].co[1] >= 0. && x[i].co[1] <= 1.);
        assert(x[i].co[2] >= 0. && x[i].co[2] <= 1.);
#endif
        /* convert x,y,z coordinates to integers in range [0,IMAX] */
        c[0] = (UINT)(x[i].co[0] * (double)IMAX);        /* x */
        c[1] = (UINT)(x[i].co[1] * (double)IMAX);        /* y */
        c[2] = (UINT)(x[i].co[2] * (double)IMAX);        /* z */
        
        /* r = log2(max(x, y, z)) */
        r = c[0];
        if (r < c[1]) r = c[1];
        if (r < c[2]) r = c[2];

        c[0] <<= 2;
        c[1] <<= 1;
        r = LOG2(r);

        /* use state tables to convert nested quadrant's 
         * coordinates level by level */
        EL = r + 1;
        r = MAXLEVEL - EL;
        key[0] = key[1] = key[2] = 0;
        state = r % 2;

        for (level = r; level < MAXLEVEL; level++) {
            /* extract 3 bits at current level */
            EL--;
            temp = ((c[0] >> (EL)) & 4)
                  |((c[1] >> (EL)) & 2)
                  |((c[2] >> (EL)) & 1);

            /* treat key[] as long shift register */
            /* shift in converted coordinate */
            /* every key[] has thirty effective bit */
            /* the last 30 are effective */
            /* the last tree bits are effective */
            key[0] = (key[0] << 3) | ((key[1] >> 27) & 7);
            key[1] = (key[1] << 3) | ((key[2] >> 27) & 7);
            key[2] = (key[2] << 3) | *(d[state] + temp);

            state = *(s[state] + temp);
        }
        
        key[0] = key[0] & EfBit;
        key[1] = key[1] & EfBit;
        key[2] = key[2] & EfBit;

        /* convert 2 part Hilbert key to double */
        x[i].sfc =  ldexp((double)key[0], k0) 
                  + ldexp((double)key[1], k1);
                  + ldexp((double)key[2], k2);
    }

    return (i == n);
}
#endif

BOOLEAN
phgInvMortonSFC(SFC *x, INT n)
{
    INT level, EL;
    UINT key[3], c[3], temp;
    INT i;

    static UINT  IMAX;
    static UINT EfBit;
    static BOOLEAN initialized = FALSE;
    static INT k0 = 0, k1 = 0, k2 = 0;

    if (!initialized) {
        initialized = TRUE;

        /* Init IMAX such that all effective bits are 1 */
        /* IMAX = 2^32) - 1 */
        IMAX = 4294967295U;

        /* 30 effective bits, all 1 */
        /* EfBit = 2^30 - 1 */
        EfBit = IMAX >> 2;

        k0 = 60 - MAXLEVEL * 3;
        k1 = 30 - MAXLEVEL * 3;
        k2 = - MAXLEVEL * 3;
    }

    for (i = 0; i < n; i++) {
        /* verification coordinats */
#if 0
        assert(x[i].co[0] >= 0. && x[i].co[0] <= 1.);
        assert(x[i].co[1] >= 0. && x[i].co[1] <= 1.);
        assert(x[i].co[2] >= 0. && x[i].co[2] <= 1.);
#endif
        /* convert x,y,z coordinates to integers in range [0,IMAX] */
        c[0] = (UINT)(x[i].co[0] * (double)IMAX);        /* x */
        c[1] = (UINT)(x[i].co[1] * (double)IMAX);        /* y */
        c[2] = (UINT)(x[i].co[2] * (double)IMAX);        /* z */
        c[0] >>= 2;
        c[1] >>= 1;

        /* use state tables to convert nested quadrant's 
         * coordinates level by level */
        key[0] = key[1] = key[2] = 0;
        EL = 30;
        for (level = 0; level < MAXLEVEL; level++) {
            /* extract 3 bits at current level */
            EL--;
            temp = ((c[0] >> (EL)) & 1)
                  |((c[1] >> (EL)) & 2)
                  |((c[2] >> (EL)) & 4);

            /* treat key[] as long shift register */
            /* shift in converted coordinate */
            /* every key[] has thirty effective bit */
            /* the last 30 are effective */
            key[0] = (key[0] << 3) | ((key[1] >> 27) & 7);
            key[1] = (key[1] << 3) | ((key[2] >> 27) & 7);
            key[2] = (key[2] << 3) | (temp);

        }
        
        key[0] = key[0] & EfBit;
        key[1] = key[1] & EfBit;
        key[2] = key[2] & EfBit;

        /* convert 2 part Hilbert key to double */
        x[i].sfc =  ldexp((double)key[0], k0) 
                  + ldexp((double)key[1], k1);
                  + ldexp((double)key[2], k2);
    }

    return (i == n);
}

#if USE_MPI

static const char *PSFCs[] = { "hilbert", "morton", NULL };
enum {HILBERT = 0, MORTON = 1};
static int PSFC = 0;

#define HSFC_EPSILON 1.6e-7
/* Hilbert space-filling curve partitioner */
BOOLEAN
phgPartitionSFC(GRID *g, INT nprocs, DOF *weights, FLOAT power)
{
    DOTS *dots;
    SIMPLEX *e;
    BOOLEAN ret;
    SFC *x = NULL;
    FLOAT out[6];       /* min and max coordinats */
    FLOAT ext[3];       /* scaling factor */
    FLOAT temp;
    FLOAT lif;
    INT i, nleaf;
    MPI_Comm comm;      /* communicator that g->nleaf > 0 */
    FLOAT x0, y0, z0;
    static const FLOAT lam[] = {0.25, 0.25, 0.25, 0.25};
    static BOOLEAN initialized = FALSE;

    if (!initialized) {
        initialized = TRUE;
        phgOptionsRegisterTitle("SFC options:", NULL, "partition");
        phgOptionsRegisterKeyword("-sfc_method", "SFC method",
                PSFCs, &PSFC);
        return FALSE;
    }

    ret = FALSE;
    if (nprocs == 1 && g->nprocs == 1)
        return ret;
    if (g->nprocs <= phgRank)
        return ret;

    /* Communicator  consisting of procs with g->nleaf != 0 */
#if NO_NEW_COMMUNICATOR
    comm = g->comm;
#else /* NO_NEW_COMMUNICATOR */
    MPI_Comm_split(g->comm, g->nleaf == 0 ? 0 : 1, 0, &comm);
#endif /* NO_NEW_COMMUNICATOR */

    /* allocate space for SFC */
    nleaf = g->nleaf;
    x = (SFC *)phgAlloc(nleaf * sizeof(SFC));
    if (nleaf > 0 && x == NULL) {
#if !NO_NEW_COMMUNICATOR
        MPI_Comm_free(&comm);
#endif /* !NO_NEW_COMMUNICATOR */
        phgError(1, "can't allocate spcae: %s  %d\n", __FILE__, __LINE__);
        return ret;
    }

    /* calculate coordinates */
    if (phgVerbosity > 0) {
        phgPrintf("SFC: init objects coordinates ... ");
    }
    i = 0;
    ForAllElements(g, e) {
        phgGeomLambda2XYZ(g, e, lam, &x0, &y0, &z0);
        x[i].co[0] = x0;          /* x */
        x[i].co[1] = y0;          /* y */
        x[i].co[2] = z0;          /* z */

        i++;
    }  /* scan all leaves */

    /* bounding box */
    out[0] = g->bbox[0][0];
    out[1] = g->bbox[0][1];
    out[2] = g->bbox[0][2];
    out[3] = g->bbox[1][0];
    out[4] = g->bbox[1][1];
    out[5] = g->bbox[1][2];

    /* enlarge bounding box */
    temp = (out[3] - out[0]) * HSFC_EPSILON;
    out[3] += temp;
    out[0] -= temp;
    temp = (out[4] - out[1]) * HSFC_EPSILON;
    out[4] += temp;
    out[1] -= temp;
    temp = (out[5] - out[2]) * HSFC_EPSILON;
    out[5] += temp;
    out[2] -= temp;

    /* scaling coordinates, such that all coordinates belong to (0, 1)^3 */
    ext[0] = out[3] - out[0]; /* length of x axis of bounding box */
    ext[1] = out[4] - out[1]; /* length of y axis of bounding box */
    ext[2] = out[5] - out[2]; /* length of z axis of bounding box */

    /* preserve domain ratio */
#if 1
    if (ext[0] < ext[1])  ext[0] = ext[1];
    if (ext[0] < ext[2])  ext[0] = ext[2];

    if (ext[0] == 0.) ext[0] = 1.;
    ext[0] = 1. / ext[0];
    ext[1] =  ext[0];
    ext[2] =  ext[0];
#else
    /* NOT preserve domain ratio */
    ext[0] = 1. / ext[0];
    ext[1] = 1. / ext[1];
    ext[2] = 1. / ext[2];
#endif

    for (i = 0; i < nleaf; i++) {
        x[i].co[0] = (x[i].co[0] - out[0]) * ext[0];
        x[i].co[1] = (x[i].co[1] - out[1]) * ext[1];
        x[i].co[2] = (x[i].co[2] - out[2]) * ext[2];
    }  /* coordinates \in [0, 1)^3 */

    if (phgVerbosity > 0) {
        phgPrintf("ok\n");
    }
    /* generate inverse SFC */
    if (PSFC == HILBERT) {
        if (phgVerbosity > 0) {
            phgPrintf("SFC: Hilbert SFC\n");
        }
        ret = phgInvHilbertSFC(x, nleaf);
    }
    else if (PSFC == MORTON) {
        if (phgVerbosity > 0) {
            phgPrintf("SFC: Morton SFC\n");
        }
        ret = phgInvMortonSFC(x, nleaf);
    }

    if (!ret) {
#if !NO_NEW_COMMUNICATOR
        MPI_Comm_free(&comm);
#endif /* !NO_NEW_COMMUNICATOR */
        phgFree(x);
        phgError(1, "can't generate inverse SFC\n");
        return ret;
    }
    if (phgVerbosity > 0) {
        phgPrintf("SFC: generate inverse SFC ... ok\n");
    }
    
    /* init dots */
    dots = (DOTS *)phgAlloc(nleaf * sizeof(DOTS));
    if (nleaf > 0 && dots == NULL) {
#if !NO_NEW_COMMUNICATOR
        MPI_Comm_free(&comm);
#endif /* !NO_NEW_COMMUNICATOR */
        phgFree(x);
        phgError(1, "can't allocate spcae: %s  %d\n", __FILE__, __LINE__);
        return ret;
    }
    i = 0;
    ForAllElements(g, e) {
        dots[i].key = x[i].sfc;
        i++;
    }
    phgFree(x);

    /* weighted condition */
    if (weights != NULL && power != 0.) {
        FLOAT wtp = -1e10, wmax;
        i = 0;
        ForAllElements(g, e) {
            dots[i].w = Pow(*DofElementData(weights, e->index), power);
            if (dots[i].w > wtp) wtp = dots[i].w;
            i++;
        }
        /* standardize weights */
        MPI_Allreduce(&wtp, &wmax, 1, PHG_MPI_FLOAT, MPI_MAX, comm);
        if (wmax > 0.) {
            wmax = 1. / wmax * 100.;
            for (i = 0; i < nleaf; i++) {
                dots[i].w *= wmax;
            }
        }
        else {
            for (i = 0; i < nleaf; i++) {
                dots[i].w = 1.;
            }
        }
    }
    else { /* no weights DOF, all are 1. */
        for (i = 0; i < nleaf; i++) {
            dots[i].w = 1.;
        }
    }
   
    /* 1d  partition */
    if (phgVerbosity > 0) {
        phgPrintf("SFC: 1D partitioning:\n");
    }

    lif = phgPartition1D(dots, nleaf, nprocs, comm);

    /* assign */
    i = 0;
    ForAllElements(g, e) {
        e->mark = dots[i].pn;
        i++;
    }
    phgFree(dots);

    if (phgVerbosity > 2) {
        INT c[nprocs];
        INT ct[nprocs];

        for (i = 0; i < nprocs; i++) {
            c[i] = 0;
        }
        ForAllElements(g, e) {
            assert(e->mark >= 0);
            assert(e->mark < nprocs);

            c[e->mark] += 1;
        }

        MPI_Allreduce(c, ct, nprocs, PHG_MPI_INT, MPI_SUM, g->comm);
        for (i = 0; i < nprocs; i++) {
            phgPrintf("\nrank: %d, # of elements: %d", i, ct[i]);
        }
        phgPrintf("\nSFC: 1d partitioning ... ok\n");
    }

    /* remap partitions */
    if (phgVerbosity > 0) {
        phgPrintf("SFC: remapping:");
    }
    phgPartitionRemap(g, nprocs, comm);
    if (phgVerbosity > 2) {
        INT c[nprocs];
        INT ct[nprocs];

        for (i = 0; i < nprocs; i++) {
            c[i] = 0;
        }
        ForAllElements(g, e) {
            assert(e->mark >= 0);
            assert(e->mark < nprocs);

            c[e->mark] += 1;
        }

        MPI_Allreduce(c, ct, nprocs, PHG_MPI_INT, MPI_SUM, g->comm);
        for (i = 0; i < nprocs; i++) {
            phgPrintf("\nrank: %d, # of elements: %d", i, ct[i]);
        }
        phgPrintf("\nSFC: remap partitions ... ok\n");
    }

#if !NO_NEW_COMMUNICATOR
    MPI_Comm_free(&comm);
#endif /* !NO_NEW_COMMUNICATOR */

    /* update information */
    i = 0;
    ForAllElements(g, e) {
        if (e->mark == phgRank) i++;
    }
    if (i == g->nleaf) {
        nleaf = 0;
    }
    else {
        nleaf = 1;
    }
    
    MPI_Allreduce(&nleaf, &i, 1, PHG_MPI_INT, MPI_SUM, g->comm);
    
    if (i == 0) { /* grid unchanged */
        ret = FALSE;
	if (phgVerbosity > 0)
            phgPrintf("%s: ** Grid Unchanged **\n", __func__);
    }
    else {        /* grid changed */
        ret = TRUE;
        g->lif = lif;
    }

    return ret;
}

#endif
